#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/uio.h>

#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include "my_index.h"
#include "Message.h"
#include "mes_func.h"
#include "protcli.h"
#include "error.h"
#include "wrap.h"
#include "my_time.h"
#include "ctrlsnmpd.h"

int
print_arcs(OBJECT_IDENTIFIER_t *oid) {
		unsigned long fixed_arcs[10];	
		unsigned long *arcs = fixed_arcs;
		int arc_type_size = sizeof(fixed_arcs[0]);	
		int arc_slots = sizeof(fixed_arcs)/sizeof(fixed_arcs[0]);
		int count;	
		int i;

		count = OBJECT_IDENTIFIER_get_arcs(oid, arcs,
			arc_type_size, arc_slots);
		if(count > arc_slots) {
			arc_slots = count;
			arcs = malloc(arc_type_size * arc_slots);
			if  (!arcs) 
			    return(-1);
			count = OBJECT_IDENTIFIER_get_arcs(oid, arcs,
				arc_type_size, arc_slots);
			if  (count != arc_slots)
			    return(-1);
		}
 		for(i = 0; i < count; i++) {
//			printf(".%lu", arcs[i]);
		}
		if(arcs != fixed_arcs) free(arcs);
		return 0;
}

message_entry_p
new_queue_msg(mesg_data_p mes,snmp_user_index_p backfire) {
    message_entry_p new;
	
    new=(message_entry_p)Calloc(1,sizeof(message_entry));
    printf("Queuing new message %i current message is NULL\n",new);
    new->message=mes;
    new->previous=new;
    new->next=new;
    new->status=MSG_READY;
    new->conseq_to=0;
    new->backfire=backfire;
    return new;
}

int
queue_msg(mesg_data_p mes,message_entry_p current,snmp_user_index_p backfire) {
    message_entry_p new;
    message_entry_p storage;

    storage=current->previous;
    new=(message_entry_p)Calloc(1,sizeof(message_entry));
//    printf("Queuing new message %i current message is %i current previous is %i\n",new,current,storage);
    new->message=mes;
    new->previous=storage;
    storage->next=new;
    new->next=current;
    current->previous=new;
//    printf("Previous for current is %i and next is %i\n",current->previous,current->next);
//    printf("Previous for new is %i and next is %i\n",new->previous,new->next);
//    printf("Previous for storage is %i and next is %i\n",storage->previous,storage->next);
    new->status=MSG_READY;
    new->conseq_to=0;
    new->backfire=backfire;
    return 0;

}

message_entry_p
free_queue_element(message_entry_p current) {
    message_entry_p new;

//    printf("Killing message %i. Prev is %i Next is %i\n",current, current->previous,current->next);
    current->next->previous=current->previous;
    current->previous->next=current->next;
    new=current->next;
//    printf("Next is %i it's previous is %i \n",new,new->previous);
//    printf("Previous %i it's next is %i\n",current->previous,current->previous->next);
    if (new==current) {
//	printf ("The queue has ended\n");
	free_mesg_data(current->message);
	free(current);
	return NULL;
    }
    free_mesg_data(current->message);
    free(current);
    return new;
}

int
send_pdu(int sock,message_entry_p current) {
    mesg_data_p info;
    Message_t *mesg;
    asn_enc_rval_t ec;
    int result, result2;
    char *outputbuf;
    struct sockaddr_in addr;

    info=(mesg_data_p)current->message;
    mesg=(Message_t *)Calloc(1,sizeof(Message_t));
    result=fill_mesg(mesg,info);
    ec = der_encode(&asn_DEF_Message, mesg, 0, 0);
    result=ec.encoded;
    result2=result;
    if (result <= 0) {
        err_msg("Encoding failure");
	return -1;
    }
    outputbuf =(char *)Malloc(result);
    ec = der_encode_to_buffer(&asn_DEF_Message, mesg, (void *) outputbuf, &result);
    if (ec.encoded != result2) {
	err_msg("Encoding failure 2");
	return -1;
    }
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(161);
    addr.sin_addr=info->ip;
    result2 = sendto(sock, outputbuf, result, 0,  (struct sockaddr *)&addr, sizeof(addr));
    free(outputbuf);
    free_message(mesg);
    return 0;
}

int
read_pdu(int sock,my_index_p index, message_entry_p * current) {
    Message_t *mesg;
    mesg_data_p new_msg;
    asn_dec_rval_t dc;
    char *buf;
    int n;
    int request_id;
    int l,m;
    int number=0;
    struct sockaddr_in ip_device;
    socklen_t ip_device_len;
    char *buf_to_send;
    int	buf_to_send_size;

    buf=(char *)Malloc(MAX_SNMP_SIZE);
//    mesg=(Message_t *)Calloc(1,sizeof(Message_t));
    while ((n=recvfrom(sock, buf, MAX_SNMP_SIZE,0,(struct sockaddr *)&ip_device,&ip_device_len)) > 0) {
	mesg=(Message_t *)Calloc(1,sizeof(Message_t));
	dc = asn_DEF_Message.ber_decoder(0, &asn_DEF_Message, (void **) &mesg, buf, n,0);
	if (dc.code != RC_OK) {
            err_msg("Decoding failure");
	    free_message(mesg);
	    continue;
        }
	else {
	    new_msg=(mesg_data_p)Calloc(1,sizeof(mesg_data));
	    l=putrefy_mesg(mesg,new_msg);
	    new_msg->ip=ip_device.sin_addr;
	    request_id=new_msg->r_id;
	    m=compare_mesg(new_msg,((message_entry_p)index->root[request_id])->message);
	    if ((compare_mesg(new_msg,((message_entry_p)index->root[request_id])->message))==0){
    	        number++;
		buf_to_send=prepare_request(new_msg,&buf_to_send_size);
		Write(((message_entry_p)index->root[request_id])->backfire->socket, buf_to_send, buf_to_send_size);
		free(buf_to_send);
	//	free_mesg_data(((message_entry_p)index->root[request_id])->message);
		my_index_free_pointer(index, request_id);
	        *current=(free_queue_element((message_entry_p)index->root[request_id]));
	    }
	    free_mesg_data(new_msg);
	}
	free_message(mesg);
    }
    free(buf);
//    printf("Nixt %i\n",number);
    return number;
}
int
send_failure(my_index_p index, message_entry_p * current, int request_id) {
    char *buf_to_send;	
    int	buf_to_send_size;

    buf_to_send=prepare_request_failure(&buf_to_send_size);
    Write(((message_entry_p)index->root[request_id])->backfire->socket, buf_to_send, buf_to_send_size);
    free(buf_to_send);
    my_index_free_pointer(index, request_id);
    *current=(free_queue_element((message_entry_p)index->root[request_id]));
}


int
main(int argc, char *argv[]) {
    int kq;
    int q;
    int request_sock,device_sock;
    int num_events;
    snmp_user_index_p index=NULL;
    int max_messages=1024;
    my_index_p message_index;
    int max_sock=64;
    struct kevent *ke,*ev;
    struct sockaddr_in *device_addr;
    struct sockaddr_un *serv;
    int backlog=5;
    struct timespec *zerotout;
    int opt_val=1;
    int *optval;
    int usocket;
    char *buf;
    int my_buf;
    mgvp mygvp=NULL;
    char *comm="private"; /*Should also update commlen*/
    int commlen=7;
    const char *path_to_socket="/tmp/ctrlsnmpd_socket";
    mesg_data_p newmsg;
    int queue_len=0;
    struct timeval *past;
    message_entry_p current;
    long timeout=1000000;
    int *len;

    signal(SIGPIPE, SIG_IGN);
    if (daemon(0,0)==-1)
	err_sys("couldn't daemonize");
    kq=Kqueue();
    buf=(char *)Malloc(sizeof(char)*MAX_MSG);
    message_index=my_index_create(max_messages);
    past=(struct timeval*)Malloc(sizeof(struct timeval));
    zerotout=(struct timespec*)Calloc(1,sizeof(struct timespec));
    current=(message_entry_p)Calloc(1,sizeof(message_entry));
    optval=&opt_val;
    unlink(path_to_socket);
    mygvp=(mgvp) Calloc(1,sizeof(mgv));
    mygvp->version=1;
    mygvp->community.buf=(uint8_t *)Malloc(sizeof(char)*commlen);
    if ((strncpy(mygvp->community.buf,comm,commlen))==NULL)
	err_quit("Unable to copy default comunity");
    mygvp->community.size=commlen;
    request_sock=Socket(AF_LOCAL,SOCK_STREAM,0);
    serv=(struct sockaddr_un *)Calloc(1,sizeof(struct sockaddr_un));
    serv->sun_family=AF_LOCAL;
    strncpy(serv->sun_path,path_to_socket,strlen(path_to_socket));
    Bind(request_sock,(struct sockaddr *)serv,sizeof(struct sockaddr_un));
    Listen(request_sock,backlog);
    ke=(struct kevent *)Calloc(1,sizeof(struct kevent));
    ev=(struct kevent *)Calloc(1,sizeof(struct kevent));
    EV_SET(ke,request_sock,EVFILT_READ, EV_ADD,0,0,NULL);
    Kevent(kq,ke,1,NULL,0,zerotout);
    device_sock=Socket(AF_INET,SOCK_DGRAM,0);
    fcntl(device_sock,F_SETFL,O_NONBLOCK);
    device_addr=(struct sockaddr_in *)Calloc(1,sizeof(struct sockaddr_in));
    device_addr->sin_family=AF_INET;
    device_addr->sin_port = htons(0);
    if (!inet_aton(argv[1],&device_addr->sin_addr))
	err_sys("bad ip");
    Bind(device_sock,(struct sockaddr *)device_addr,sizeof(struct sockaddr_in));
    bzero(ke,sizeof(struct kevent));
    EV_SET(ke,device_sock,EVFILT_READ, EV_ADD,0,0,NULL);
    Kevent(kq,ke,1,NULL,0,zerotout);
    index=(snmp_user_index_p)Calloc(max_sock,sizeof(struct snmp_user_index));
    while(1) {
	if (queue_len) {
	    num_events = Kevent(kq,NULL,0,ev,1,zerotout);
	}
	else {
	    num_events = Kevent(kq,NULL,0,ev,1,NULL);
	    err_msg("message_index %i",message_index->free);
	}
	if (num_events > 0) {
	    /* receiving is of more importance */
	    if (ev->ident==request_sock) {
		/* new connection */
		usocket=Accept(request_sock,NULL,NULL);
		if (VFLAG)
		    err_msg("accepted connection %i",usocket);
		bzero(ke,sizeof(struct kevent));
		EV_SET(ke,usocket,EVFILT_READ, EV_ADD,0,0,NULL);		
		Kevent(kq,ke,1,NULL,0,zerotout);
		if (usocket>=max_sock) {
		    max_sock*=2;
		    index=(snmp_user_index_p)Reallocf((void*)index,max_sock*sizeof(struct snmp_user_index));
		    err_msg("Reallocating index for clients");
		}
		index[usocket].type=1;
		index[usocket].socket=usocket;
	    }
	    else if (ev->ident==device_sock) {
		/* receive something from device */
		if (VFLAG)
		    err_msg("received from device socket");
		queue_len-=read_pdu(ev->ident,message_index,&current);
	    }
	    else {
		/* received something from client */
		if (VFLAG)
		    err_msg("received new task from client");
		if (index[ev->ident].type==1) {
		    if ((my_buf=read(ev->ident, buf, MAX_MSG)) > 0){
		        newmsg=proceed_request(buf,my_buf,mygvp);
			if (!queue_len) {
			    current=new_queue_msg(newmsg,&index[ev->ident]);
			    queue_len++;
			}
			else {
			    queue_msg(newmsg,current,&index[ev->ident]);
			    queue_len++;
			}
		    }
		    else {
			err_msg("Zero bytes received from client in usocket %i",ev->ident);
			bzero(ke,sizeof(struct kevent));
			EV_SET(ke,ev->ident,EVFILT_READ, EV_DELETE,0,0,NULL);
			Kevent(kq,ke,1,NULL,0,zerotout);
			index[ev->ident].type=0;
			close(ev->ident);
		    }
		}
		else {
		    err_msg("Came an answer to unexpected client");
		}
	    }
	}
	else if (queue_len>0){
	    /* here we can send anything we like */
	    switch (current->status) {
		case MSG_READY:
		    current->message->r_id=my_index_add_to(message_index,(void *)current);
		    q=send_pdu(device_sock,current);
		    if (q == 0){
			if ((gettimeofday(past, NULL)) != 0)
		            err_quit("timer failed");
		        current->sent.tv_sec=past->tv_sec;
		        current->sent.tv_usec=past->tv_usec;
		        current->status=MSG_SENT;
		        current=current->next;
		    }
		    else {
			/*Just say that it was an error to client and REMOVE entry from list*/
			err_msg("Sending failed");
			send_failure(message_index,&current,current->message->r_id);
			queue_len--;
		    }
		break;
		case MSG_SENT:
	            if (check_time(&(current->sent),timeout,1) > 0){
			current->status=MSG_READY;
			current->conseq_to++;
			if (current->conseq_to>MAXTO) {
			    err_msg("Timeout > %i times. Giving up",MAXTO);
			    send_failure(message_index,&current,current->message->r_id);
			    queue_len--;
			}
		    }
		    else {
		        current=current->next;
		    }
		break;
		
	    }
	}
	else {
	    /*nothing to receive and nothing to send*/
	}
    }
}