#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/uio.h>

#include <stdlib.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <mife.h>
#include <babolo/parser.h>

#include "error.h"
#include "wrap.h"
#include "Message.h"
#include "mes_func.h"
#include "protcli.h"
#include "swconf.h"

OBJECT_IDENTIFIER_t *
my_ether_atooid(char *str){
	char *position;
	int ether[6];
	enum {
		ST_WAITLIM,
		ST_WAITDIGIT,	
		ST_DIGIT,
	} state = ST_WAITDIGIT;
	int value;
	int number=0;
	OBJECT_IDENTIFIER_t *result;
	uint8_t *buf;
	int i;

	for(position=str;position<str+strlen(str);position++) {
		switch(*position) {
			case '1': case '2': case '3' :case'4':
			case '5': case '6': case '7' :case'8':
			case '9': case '0':{
				if (state==ST_WAITDIGIT) {
					state=ST_DIGIT;
					value=(*position)-'0';
				} else if (state==ST_DIGIT){
					value=value*16+(*position)-'0';
					state=ST_WAITLIM;
				}
				else{
					return NULL;
				}
				break;
			}
			case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':{
				if (state==ST_WAITDIGIT) {
					state=ST_DIGIT;
					value=(*position)-'a'+10;
				} else if (state==ST_DIGIT){
					value=value*16+(*position)-'a'+10;
					state=ST_WAITLIM;
				}
				else{
					return NULL;
				}
				break;

			}
			case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':{
				if (state==ST_WAITDIGIT) {
					state=ST_DIGIT;
					value=(*position)-'A'+10;
				} else if (state==ST_DIGIT){
					value=value*16+(*position)-'A'+10;
					state=ST_WAITLIM;
				}
				else{
					return NULL;
				}
				break;
			}
			case ':': case '-':{
				
				if (state==ST_WAITLIM) {
					if (number>4)
						return NULL;
					ether[number]=value;
					number++;
					state=ST_WAITDIGIT;
				}
				else
					return NULL;
				break;
			}
			default:
				return NULL;
		}
	}
	if (number!=5)
		return NULL;
	result=(OBJECT_IDENTIFIER_t *)Malloc(sizeof(OBJECT_IDENTIFIER_t));
	buf=(uint8_t *)Malloc(sizeof(uint8_t)*6*2);
	result->buf=buf;
	ether[number]=value;
	for (number=0;number<6;number++){
		i=OBJECT_IDENTIFIER_set_single_arc(buf,&ether[number],sizeof(long),0);
		buf+=i;
	}
	result->size=buf-result->buf;
	return result;
}

list_entry_p
fdb_st_get(struct babolo_parm *bp,sw_inf_p sw_info,list_entry_p *res) {
	list_entry_p result=*res;
	fdb_snmp_p current;	
	row_snmp_data_p rd_place;
	vb_data_p vb;
	int port;
	int vlan;
	OBJECT_IDENTIFIER_t *tmp,*vl;
	OBJECT_IDENTIFIER_t *mac;
	char *end_ptr;
	int suffix=0;
	int port_group;
	int arg=3;

	if (bp->argc != 3) {
		err_msg("Bad static fdb string");
		return NULL;
	}
	port=(int)strtol(bp->argv[0],&end_ptr,10);
	if (((end_ptr) && *end_ptr!=0) || port > sw_info->num_ports || port < 1) {
		err_msg("Couldn't parse port number in fdb");
		return NULL;
	}
	printf("Port is\n");
	vlan=(int)strtol(bp->argv[1],&end_ptr,10);
	if (((end_ptr) && *end_ptr!=0) || vlan > 4095 || vlan < 1) {
		err_msg("Couldn't parse vlan number in fdb");
		return NULL;
	}
	if ((mac=my_ether_atooid(bp->argv[2]))==NULL){
		err_msg("Couldn't parse mac address in fdb");
		return NULL;
	}
	vl=(OBJECT_IDENTIFIER_t *)Malloc(sizeof(OBJECT_IDENTIFIER_t));
	vl->buf=(uint8_t *)Malloc(sizeof(long));
	if ((vl->size=OBJECT_IDENTIFIER_set_single_arc(vl->buf,&vlan,sizeof(long),0))<=0){
		free(vl->buf);
		free(vl);
		err_msg("Unable to parse vlan integer into oid part\n");
		return 0;
	}
	tmp=(OBJECT_IDENTIFIER_t *)Malloc(sizeof(OBJECT_IDENTIFIER_t));
	concatenate_oid(tmp,vl,mac);
	if ((vl->size=OBJECT_IDENTIFIER_set_single_arc(vl->buf,&suffix,sizeof(long),0))<=0){
		free(tmp->buf);
		free(tmp);
		free(vl->buf);
		free(vl);
		err_msg("Unable to parse suffix integer into oid part\n");
	}
	current=(fdb_snmp_p)Malloc(sizeof(fdb_snmp_s));
	current->flags=0;
	current->row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
	current->row->row_begin=&sw_info->fdb_st_row_oid;
	concatenate_oid(&current->row->tag_row_id,tmp,vl);
	free(tmp->buf);
	free(tmp);
	free(vl->buf);
	free(vl);
	free(mac->buf);
	free(mac);
	current->fdb=&current->row->tag_row_id;
	rd_place=current->row->row_data=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
	rd_place->suffix=&sw_info->fdb_st_port_suffix;
	rd_place->data=vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
	vb->vbv_present=VarBindValue_PR_value;
    	vb->os_present=ObjectSyntax_PR_simple;
	vb->st_present=SimpleSyntax_PR_string_value;
	vb->value.size=sw_info->num_ports/8+1;
	vb->value.buf=(uint8_t *)Calloc(vb->value.size,sizeof(uint8_t));
	port_group=port/8;
	if (port_group * 8 == port)
		port_group--;
	vb->value.buf[port_group]=0x80 >> (port-port_group*8-1);
	rd_place->next=(row_snmp_data_p)Calloc(1,sizeof(row_snmp_data_s));
	rd_place=rd_place->next;
	rd_place->suffix=&sw_info->fdb_st_state_suffix;
	rd_place->data=vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
	vb->vbv_present=VarBindValue_PR_value;
    	vb->os_present=ObjectSyntax_PR_simple;
	vb->st_present=SimpleSyntax_PR_integer_value;
	vb->value.size=sizeof(long);
	vb->value.buf=(uint8_t *)Malloc(vb->value.size);
	bcopy(&arg,vb->value.buf,vb->value.size);
	rd_place->next=NULL;
	result->entry=(void*)current;
	result->type=FDB_ST_ENTRY;
	result->flags=0;
	result->next=(list_entry_p)Malloc(sizeof(list_entry));
	return result->next;

}

list_entry_p
port_get(struct babolo_parm *bp,sw_inf_p sw_info,list_entry_p *res) {
	list_entry_p result=*res;
	port_snmp_p current;
	row_snmp_data_p rd_place;
	vb_data_p vb;

	char *end_ptr;
	int port;
	int up;
	int speed;
	int arg;	

	int i,j;

	if (bp->argc != 3) {
		err_msg("Bad port string");
		return NULL;
	}
	port=(int)strtol(bp->argv[0],&end_ptr,10);
	if (((end_ptr) && *end_ptr!=0) || port > sw_info->num_ports || port < 1) {
		err_msg("Couldn't parse port number");
		return NULL;
	}
	up=(int)strtol(bp->argv[1],&end_ptr,10);
	if (((end_ptr) && *end_ptr!=0) || up > 3 || up < 2) {
		err_msg("Only 2(down) or 3 (up) required");
		return NULL;
	}
	speed=(int)strtol(bp->argv[2],&end_ptr,10);
	if (((end_ptr) && *end_ptr!=0) || speed > 8 || speed < 2) {
		err_msg("Only 2 to 8 enabled for port speed");
		return NULL;
	}
	current=(port_snmp_p)Malloc(sizeof(port_snmp_s));
	current->port=port;
	current->flags=0;
	current->row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
	current->row->row_begin=&sw_info->port_row_oid;
	current->row->tag_row_id.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	current->row->tag_row_id.size=OBJECT_IDENTIFIER_set_single_arc(current->row->tag_row_id.buf, &(current->port), sizeof(long),0);
	rd_place=current->row->row_data=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
	for (i=1;i<3;i++) {
		switch(i) {
			case 1: {
				rd_place->suffix=&sw_info->port_state_suffix;
				break;
			}
			case 2: {
				rd_place->suffix=&sw_info->port_speed_suffix;
				break;
			}
		}
		rd_place->data=vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
		vb->vbv_present=VarBindValue_PR_value;
    		vb->os_present=ObjectSyntax_PR_simple;
		vb->st_present=SimpleSyntax_PR_integer_value;
		vb->value.size=sizeof(long);
		vb->value.buf=(uint8_t *)Malloc(vb->value.size);
		arg=strtol(bp->argv[i],&end_ptr,10);
		bcopy(&arg,vb->value.buf,vb->value.size);
		if (i<2) {
			rd_place->next=(row_snmp_data_p)Calloc(1,sizeof(row_snmp_data_s));
			rd_place=rd_place->next;
		}
	}
	rd_place->next=NULL;
	result->entry=(void*)current;
	result->type=PORT_ENTRY;
	result->flags=0;
	result->next=(list_entry_p)Malloc(sizeof(list_entry));
	return result->next;
}

list_entry_p
vlan_get(struct babolo_parm *bp,sw_inf_p sw_info,list_entry_p *res) {

	list_entry_p result=*res;
	vl_snmp_p current;
	row_snmp_data_p rd_place;

	vb_data_p vb;

	char *end_ptr;
	int tag;
	int i,j;
	uint8_t testing;
	int counter,counter_slot;

	if (bp->argc == 5) {
		/* initial filling in */
		tag=(int)strtol(bp->argv[0],&end_ptr,10);
		if (((end_ptr) && *end_ptr!=0) || (tag<1) || (tag>4094)) {
			err_msg("Couldn't parse vlan integer");
			return NULL;
	        }
		for (i=1;i<4; i++) {
			if (strlen(bp->argv[i])!=sw_info->num_ports) {
				err_msg("Port string size differs. Line Ignored");
				return NULL;
			}
		}
		current=(vl_snmp_p)Malloc(sizeof(vl_snmp_s));
		current->vid=tag;
		current->flags=0;
		current->row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
		current->row->row_begin=&sw_info->vlan_row_oid;
		current->row->tag_row_id.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	        current->row->tag_row_id.size=OBJECT_IDENTIFIER_set_single_arc(current->row->tag_row_id.buf, &(current->vid), sizeof(long),0);
		rd_place=current->row->row_data=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
		for (i=1;i<4;i++) {
			switch (i) {
				case 1: {
					rd_place->suffix=&sw_info->vlan_egress_suffix;
					break;
				}
				case 2: {
					rd_place->suffix=&sw_info->vlan_forbidden_suffix;
					break;
				}
				case 3: {
					rd_place->suffix=&sw_info->vlan_untagged_suffix;
					break;
				}
			}
			rd_place->data=vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
			vb->vbv_present=VarBindValue_PR_value;
    			vb->os_present=ObjectSyntax_PR_simple;
			vb->st_present=SimpleSyntax_PR_string_value;
			vb->value.size=sw_info->num_ports/8+1;
			vb->value.buf=(uint8_t *)Calloc(vb->value.size,sizeof(uint8_t));
			counter=counter_slot=0;
			for (j=0;j<sw_info->num_ports;j++){
				if (((int)*(bp->argv[i]+j)==48) || ((int)*(bp->argv[i]+j)==49)) {
		    			counter++;
		    			vb->value.buf[counter_slot]|=((int)*(bp->argv[i]+j)-48)<<(8-counter);
		    			if (counter==8) {
						counter=0;
						counter_slot++;
					}
				}
				else {
					err_msg("Unknown symbols in line. Skipping");
					free_vl_snmp_p(current);
					return NULL;
				}
			}
			switch (i) {
				case 2: {
					for (j=0;j<sw_info->num_ports/8+1;j++) {
						testing=vb->value.buf[j] & current->row->row_data->data->value.buf[j];
						if (testing) {
		    					err_msg("forbidden port %s where it is in egress %s",bp->argv[2],bp->argv[1]);
		    					current->row->row_data->data->value.buf[j]&=~vb->value.buf[j];
						}
	    				}
					break;
				}
				case 3: {
					for (j=0;j<sw_info->num_ports/8+1;j++) {
						testing=vb->value.buf[j] & (~current->row->row_data->data->value.buf[j]);
						
						if (testing) {
		    					err_msg("untagged port %s where it is not in egress %s ",bp->argv[3],bp->argv[1]);
		    					vb->value.buf[j]^=testing;
						}
						testing=sw_info->vlan_untagged_ports.buf[j] & (vb->value.buf[j]);
						if (testing) {
							err_msg("untagged ports overlapped in vlan %i",tag);
							vb->value.buf[j]^=testing;
						}
						sw_info->vlan_untagged_ports.buf[j]|=vb->value.buf[j];

					}
					break;
				}
			}
			rd_place->next=(row_snmp_data_p)Calloc(1,sizeof(row_snmp_data_s));
			rd_place=rd_place->next;
		}
		/* now the name */
		rd_place->suffix=&sw_info->vlan_name_suffix;
		rd_place->next=NULL;
		rd_place->data=vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
		vb->vbv_present=VarBindValue_PR_value;
    		vb->os_present=ObjectSyntax_PR_simple;
		vb->st_present=SimpleSyntax_PR_string_value;
		vb->value.size=strlen(bp->argv[4]);
		vb->value.buf=(uint8_t *)Malloc(sizeof(uint8_t)*vb->value.size);
		bcopy(bp->argv[4],vb->value.buf,vb->value.size);
	}
	else {
		return NULL;
	}
	result->entry=(void*)current;
	result->type=VLAN_ENTRY;
	result->flags=0;
	result->next=(list_entry_p)Malloc(sizeof(list_entry));
	return result->next;
}


int 
fdb_st_compare(void *p_void, void *q_void,int err_gen) {
	fdb_snmp_p p=(fdb_snmp_p)p_void;
	fdb_snmp_p q=(fdb_snmp_p)q_void;
	int size,i;

	if (p->fdb->size < q->fdb->size)
		size=p->fdb->size;
	else
		size=q->fdb->size;
	for (i=0;i<size;i++)
		if (p->fdb->buf[i]< q->fdb->buf[i])
			return -1;
		else if (p->fdb->buf[i] > q->fdb->buf[i])
			return 1;
	if (p->fdb->size < q->fdb->size)
		return -1;
	else if (p->fdb->size > q->fdb->size)
		return 1;
	return 0;
}

int
port_compare(void *p_void, void *q_void,int err_gen) {
	port_snmp_p p=(port_snmp_p)p_void;
	port_snmp_p q=(port_snmp_p)q_void;

	if (p->port > q->port)
		return 1;
	else if (p->port < q->port)
		return -1;
	else {
		if (err_gen)
			err_msg("Ports with equal port numbers. Result unpredictable");
		return 0;
	}
}

int
vlan_compare(void *p_void, void *q_void,int err_gen) {
	vl_snmp_p p=(vl_snmp_p)p_void;
	vl_snmp_p q=(vl_snmp_p)q_void;

	if (p->vid > q->vid){
		return 1;
	}
	else if (p->vid < q->vid) {
		return -1;
	}
	else {
		if (err_gen)
			err_msg("Vlans with equal tags. Result unpredictable");
		return 0;
	}
}


int
free_row_data_snmp_p(row_snmp_data_p tbfreed){
	if (tbfreed->next)
		free_row_data_snmp_p(tbfreed->next);
	free_vb_data(tbfreed->data);
	return 0;
}

int
free_row_snmp_p(row_snmp_p tbfreed) {
	free(tbfreed->tag_row_id.buf);
	free_row_data_snmp_p(tbfreed->row_data);
	return 0;
} 

int
free_vl_snmp_p (vl_snmp_p tbfreed) {
	free_row_snmp_p(tbfreed->row);
	free(tbfreed);
	return 0;
}


int
compare_entry(list_entry_p_base p, list_entry_p_base q){
	if ((*p)->type > (*q)->type)
	    return 1;
	else if ((*p)->type == ((*q)->type)) {
		if (par_hdlrs[(*p)->type].c_p_comp_f) {
			return (*(par_hdlrs[(*p)->type]).c_p_comp_f)((*p)->entry,(*q)->entry,1);
		}
		else {
			err_msg ("Couldn't find how to compare. Result is unpredictable");
			return 0;
		}
	}
	else
	    return -1;
}

int
compare_entry_no_err(list_entry_p_base p, list_entry_p_base q){
	if ((*p)->type > (*q)->type)
	    return 1;
	else if ((*p)->type == ((*q)->type)) {
		if (par_hdlrs[(*p)->type].c_p_comp_f) {
			return (*(par_hdlrs[(*p)->type]).c_p_comp_f)((*p)->entry,(*q)->entry,0);
		}
		else {
			return 0;
		}
	}
	else
	    return -1;
}


l_ref_p
open_task(sw_inf_p sw_info){
	mife_descriptor *in;
	int len;
	u_char *r;
	u_char *q=NULL;
	u_char *error_place;
	struct babolo_parm *bp;
	register const struct config_parameter_handler *p;
	list_entry_p cur,beg,tmp;
	l_ref_p res;
	int number=0;
	int i;

	if (!(in=mife_opef(MIFE_FULL, fileno(stdin)))) {
		err_quit("Failed to open file with task");
	}
	if ((len=mife_read(in, 0, 0)) < 0 || !(q = mife_get(in, 0))) {
		err_quit("Mife erron reading");
	}
	r=q;
	cur=(list_entry_p)Malloc(sizeof(list_entry));
	beg=cur;
	while (r < q + len){
		error_place=r;
		if ((bp = babolo_getparm(Bpars_CRLN |Bpars_LFLN, &r, ""))==NULL) {
			err_quit("Input string error : %s", error_place);
		}
		if (!bp->argc)
			continue;
		for (p = par_hdlrs; p->p_name; p++) {
			if (strcmp(*bp->argv,p->p_name) == 0)
				break;
		}
		if (p->c_p_read_f) {
			bp->argv++;
			bp->argc--;
			if ((tmp=(*p->c_p_read_f)(bp,sw_info,&cur))!=NULL){
				cur=tmp;
				number++;
			}
			else {
				err_msg("Failed to add");
			}
		}
		else if (p->p_name) {
			err_msg("Not implemented");
		}
	}
	cur->type=END_OF_LIST;
	cur->next=NULL;
	cur->flags=0;
	cur->entry=NULL;
	number++;
	res=(l_ref_p)Malloc(sizeof(l_ref));
	res->base=(list_entry_p_base)Malloc(sizeof(list_entry_p)*number);
	res->number=number;
	res->entries=0;
	cur=beg;
	for (i=0;i<number;i++) {
		(res->base[i])=cur;
		res->entries|=1<<cur->type;
		cur=cur->next;
		if ((cur==NULL) && (i<number-1))
			err_msg("Wrong pointer");
	}
	qsort((void *)res->base, res->number, sizeof(list_entry_p), (int (*)(const void *, const void *))compare_entry);
	for (i=1;i<number; i++) {
		if (compare_entry(&(res->base[i]),&(res->base[i-1]))==0) {
			(res->base[i])->flags |= NOT_PERFORM;
			err_msg("Duplicate task. Marked not to be performed");
		}
	}
	return res;	
}

int
try_fit(cl_msg_hdr_p *initial,vb_data_p vb,char *copy_place){
	int vb_len;
	cl_msg_hdr_p init= *initial;
	cl_vb_hdr_p  vb_init;


	vb_len=sizeof(cl_vb_hdr)+vb->oid.size;
	if (vb->value.size!=0)
		vb_len+=vb->value.size;
	if (init->len + vb_len < MAX_MSG) {
		init->vb_n++;
		vb_init=(cl_vb_hdr_p)copy_place;
        	vb_init->len=vb_len;
        	setVBV_POSITION(vb_init->flags,vb->vbv_present);
        	setOS_POSITION(vb_init->flags,vb->os_present);
        	setST_POSITION(vb_init->flags,vb->st_present);
        	vb_init->oid_len=vb->oid.size;
        	copy_place=(char *)(vb_init+1);
        	bcopy(vb->oid.buf,copy_place,vb->oid.size);
        	copy_place+=vb->oid.size;
		if (vb->value.size != 0){
        		vbv_cpv(copy_place,vb);
		}
		init->len+=vb_len;
	}
	else
		return 0;
	return vb_len;	
}


list_entry_p
fdb_st_get_cur(list_entry_p *result,sw_inf_p sw_info,int *num){
	list_entry_p cur=*result;
	cl_msg_hdr_p init;
	vb_data_p vb;
	char *copy_place;
	int shift;

	mesg_data_p newmsg;
	char *buf;
	OBJECT_IDENTIFIER_t *beginning;
	int i;

	fdb_snmp_p current;
	row_snmp_data_p rd_place;

	init=(cl_msg_hdr_p)Calloc(1,sizeof(char)*MAX_MSG);
	init->len=sizeof(cl_msg_hdr);

	bcopy(&(sw_info->ip),&(init->ip),sizeof(struct in_addr));
	init->pdu_pr=PDUs_PR_get_next_request;
	init->vb_n=0;
	vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
        vb->vbv_present=VarBindValue_PR_unSpecified;
	vb->os_present=0;
	vb->st_present=0;
	vb->value.size=0;
	concatenate_oid(&(vb->oid),&(sw_info->fdb_st_row_oid),&(sw_info->fdb_st_state_suffix));	
	copy_place=(char*)(init+1);
	if ((shift=try_fit(&init,vb,copy_place))==0) {
		err_msg("Couldn't construct get-next request because of too little MAX_MSG");
		free(init);
		free(vb->oid.buf);
		free(vb);
		return NULL;
	}
	i=send(sw_info->socket, init,init->len,0);
    	buf=(char *)Malloc(MAX_MSG);
    	i=recvfrom(sw_info->socket, buf, MAX_MSG,0,NULL,NULL);
   	if ((newmsg=proceed_request(buf,i,sw_info->mygvp))==NULL) {
		free(buf);
		free(vb->oid.buf);
		free(vb);
		free(init);
		err_msg("Unable to parse responce message");
		return NULL;
	}
	beginning=(OBJECT_IDENTIFIER_t *)Malloc(sizeof(OBJECT_IDENTIFIER_t));
	concatenate_oid(beginning,&(sw_info->fdb_st_row_oid),&(sw_info->fdb_st_state_suffix));
	while(!notintree(beginning,&(newmsg->data->oid))) {
		current=(fdb_snmp_p)Malloc(sizeof(fdb_snmp_s));
		current->flags=0;
		current->row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
		current->row->row_begin=&sw_info->fdb_st_row_oid;
		current->row->tag_row_id.size=newmsg->data->oid.size-beginning->size;
		current->row->tag_row_id.buf=(uint8_t *)Malloc(current->row->tag_row_id.size);
		bcopy(&(newmsg->data->oid.buf[beginning->size]),current->row->tag_row_id.buf,current->row->tag_row_id.size);
		current->row->row_data=rd_place=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
		current->fdb=&current->row->tag_row_id;
		rd_place->suffix=&sw_info->fdb_st_port_suffix;
		rd_place->next=(row_snmp_data_p)Calloc(1,sizeof(row_snmp_data_s));
		rd_place=rd_place->next;
		rd_place->suffix=&sw_info->fdb_st_state_suffix;
		rd_place->next=NULL;
		if (!get_row_data(current->row,sw_info)) {
			err_msg("Unable to retrieve vlan information");
			free(current);
			free(cur);
			free(buf);
			free(vb->oid.buf);
			free(vb);
			free(init);
			return NULL;
		}
		cur->entry=(void*)current;
		cur->type=FDB_ST_ENTRY;
		cur->flags=0;
		cur->next=(list_entry_p)Malloc(sizeof(list_entry));
		cur=cur->next;
		(*num)++;
		init->len=sizeof(cl_msg_hdr);
		init->vb_n=0;
		vb->oid=newmsg->data->oid;
		copy_place=(char*)(init+1);
		if ((shift=try_fit(&init,vb,copy_place))==0) {
			err_msg("Couldn't construct get-next request because of too little MAX_MSG");
			free(init);
			free(vb->oid.buf);
			free(vb);
			return NULL;
		}
		i=send(sw_info->socket, init,init->len,0);
		buf=(char *)Malloc(MAX_MSG);
		i=recvfrom(sw_info->socket, buf, MAX_MSG,0,NULL,NULL);
		if ((newmsg=proceed_request(buf,i,sw_info->mygvp))==NULL) {
			free(buf);
			free(vb->oid.buf);
			free(vb);
			free(init);
			err_msg("Unable to parse responce message");
			return NULL;
		}
	}
	return cur;
}

list_entry_p
port_get_cur(list_entry_p *result,sw_inf_p sw_info,int *num){
	list_entry_p cur=*result;
	cl_msg_hdr_p init;
	vb_data_p vb;
	char *copy_place;
	int shift;

	int i,j;

	mesg_data_p newmsg;
	char *buf;
	OBJECT_IDENTIFIER_t *beginning;

	port_snmp_p current;
	row_snmp_data_p rd_place;
	int check=0;
	uint8_t b;


	init=(cl_msg_hdr_p)Calloc(1,sizeof(char)*MAX_MSG);
	init->len=sizeof(cl_msg_hdr);

	bcopy(&(sw_info->ip),&(init->ip),sizeof(struct in_addr));
	init->pdu_pr=PDUs_PR_get_next_request;
	init->vb_n=0;
	vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
        vb->vbv_present=VarBindValue_PR_unSpecified;
	vb->os_present=0;
	vb->st_present=0;
	vb->value.size=0;
	concatenate_oid(&(vb->oid),&(sw_info->port_row_oid),&(sw_info->port_index_suffix));
	copy_place=(char*)(init+1);
	if ((shift=try_fit(&init,vb,copy_place))==0) {
		err_msg("Couldn't construct get-next request because of too little MAX_MSG");
		free(init);
		free(vb->oid.buf);
		free(vb);
		return NULL;
	}
	copy_place+=shift;
	i=send(sw_info->socket, init,init->len,0);
    	buf=(char *)Malloc(MAX_MSG);
    	i=recvfrom(sw_info->socket, buf, MAX_MSG,0,NULL,NULL);
   	if ((newmsg=proceed_request(buf,i,sw_info->mygvp))==NULL) {
		free(buf);
		free(vb->oid.buf);
		free(vb);
		free(init);
		err_msg("Unable to parse responce message");
		return NULL;
	}
	beginning=(OBJECT_IDENTIFIER_t *)Malloc(sizeof(OBJECT_IDENTIFIER_t));
	concatenate_oid(beginning,&(sw_info->port_row_oid),&(sw_info->port_index_suffix));
	while(!notintree(beginning,&(newmsg->data->oid))) {
		current=(port_snmp_p)Malloc(sizeof(port_snmp_s));
		check=0;
		for(j=beginning->size;j<newmsg->data->oid.size;j++) {	
			b = newmsg->data->oid.buf[j];
			if (b & 0x80)			/* Continuation expected */
				continue;
			if (check) {
				err_msg("Unable to parse row suffix");
				free(current);
				free(cur);
				free(buf);
				free(vb->oid.buf);
				free(vb);
				free(init);
				return NULL;
			}
			if (OBJECT_IDENTIFIER_get_single_arc(&(newmsg->data->oid.buf[beginning->size]),j-beginning->size+1,0,&(current->port), sizeof(int))) {
				err_msg("Unable to identify vlan tag");
				free(current);
				free(cur);
				free(buf);
				free(vb->oid.buf);
				free(vb);
				free(init);
				return NULL;
			}
			check=1;
		}
		current->flags=0;
		current->row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
		current->row->row_begin=&sw_info->port_row_oid;
		current->row->tag_row_id.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	        current->row->tag_row_id.size=OBJECT_IDENTIFIER_set_single_arc(current->row->tag_row_id.buf, &(current->port), sizeof(long),0);
		current->row->row_data=rd_place=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
		for (i=1;i<3;i++) {
			switch (i) {
				case 1: {
					rd_place->suffix=&sw_info->port_state_suffix;
					break;
				}
				case 2: {
					rd_place->suffix=&sw_info->port_speed_suffix;
					break;
				}
			}
			if (i!=2) {
				rd_place->next=(row_snmp_data_p)Calloc(1,sizeof(row_snmp_data_s));
				rd_place=rd_place->next;
			}
			else rd_place->next=NULL;
		}
		if (!get_row_data(current->row,sw_info)) {
			err_msg("Unable to retrieve vlan information");
			free(current);
			free(cur);
			free(buf);
			free(vb->oid.buf);
			free(vb);
			free(init);
			return NULL;
		}
		cur->entry=(void*)current;
		cur->type=PORT_ENTRY;
		cur->flags=0;
		cur->next=(list_entry_p)Malloc(sizeof(list_entry));
		cur=cur->next;
		(*num)++;
		init->len=sizeof(cl_msg_hdr);
		init->vb_n=0;
		vb->oid=newmsg->data->oid;
		copy_place=(char*)(init+1);
		if ((shift=try_fit(&init,vb,copy_place))==0) {
			err_msg("Couldn't construct get-next request because of too little MAX_MSG");
			free(init);
			free(vb->oid.buf);
			free(vb);
			return NULL;
		}
		i=send(sw_info->socket, init,init->len,0);
		buf=(char *)Malloc(MAX_MSG);
		i=recvfrom(sw_info->socket, buf, MAX_MSG,0,NULL,NULL);
		if ((newmsg=proceed_request(buf,i,sw_info->mygvp))==NULL) {
			free(buf);
			free(vb->oid.buf);
			free(vb);
			free(init);
			err_msg("Unable to parse responce message");
			return NULL;
		}
	}
	return cur;	
}

list_entry_p
vlan_get_cur(list_entry_p *result,sw_inf_p sw_info,int *num){
	list_entry_p cur=*result;
	cl_msg_hdr_p init;
	vb_data_p vb;
	char * copy_place;
	int shift;
	int i,j;
	mesg_data_p newmsg;
	char *buf;
	OBJECT_IDENTIFIER_t *beginning;
	vl_snmp_p current;
	row_snmp_data_p rd_place;
	int check=0;
	uint8_t b;

	init=(cl_msg_hdr_p)Calloc(1,sizeof(char)*MAX_MSG);
	init->len=sizeof(cl_msg_hdr);
	bcopy(&(sw_info->ip),&(init->ip),sizeof(struct in_addr));
	init->pdu_pr=PDUs_PR_get_next_request;
	init->vb_n=0;
        vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
        vb->vbv_present=VarBindValue_PR_unSpecified;
	vb->os_present=0;
	vb->st_present=0;
	vb->value.size=0;
	concatenate_oid(&(vb->oid),&(sw_info->vlan_row_oid),&(sw_info->vlan_init_suffix));
	copy_place=(char*)(init+1);
	if ((shift=try_fit(&init,vb,copy_place))==0) {
		err_msg("Couldn't construct get-next request because of too little MAX_MSG");
		free(init);
		free(vb->oid.buf);
		free(vb);
		return NULL;
	}
	copy_place+=shift;
	i=send(sw_info->socket, init,init->len,0);
    	buf=(char *)Malloc(MAX_MSG);
    	i=recvfrom(sw_info->socket, buf, MAX_MSG,0,NULL,NULL);
   	if ((newmsg=proceed_request(buf,i,sw_info->mygvp))==NULL) {
		free(buf);
		free(vb->oid.buf);
		free(vb);
		free(init);
		err_msg("Unable to parse responce message");
		return NULL;
	}
	beginning=(OBJECT_IDENTIFIER_t *)Malloc(sizeof(OBJECT_IDENTIFIER_t));
	concatenate_oid(beginning,&(sw_info->vlan_row_oid),&(sw_info->vlan_init_suffix));
	while(!notintree(beginning,&(newmsg->data->oid))) {
		current=(vl_snmp_p)Malloc(sizeof(vl_snmp_s));
		check=0;
		for(j=beginning->size;j<newmsg->data->oid.size;j++) {	
			b = newmsg->data->oid.buf[j];
			if (b & 0x80)			/* Continuation expected */
				continue;
			if (check) {
				err_msg("Unable to parse row suffix");
				free(current);
				free(cur);
				free(buf);
				free(vb->oid.buf);
				free(vb);
				free(init);
				return NULL;
			}
			if (OBJECT_IDENTIFIER_get_single_arc(&(newmsg->data->oid.buf[beginning->size]),j-beginning->size+1,0,&(current->vid), sizeof(int))) {
				err_msg("Unable to identify vlan tag");
				free(current);
				free(cur);
				free(buf);
				free(vb->oid.buf);
				free(vb);
				free(init);
				return NULL;
			}
			check=1;
		}
		current->flags=0;
		current->row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
		current->row->row_begin=&sw_info->vlan_row_oid;
		current->row->tag_row_id.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	        current->row->tag_row_id.size=OBJECT_IDENTIFIER_set_single_arc(current->row->tag_row_id.buf, &(current->vid), sizeof(long),0);
		current->row->row_data=rd_place=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
		for (i=1;i<5;i++) {
			switch (i) {
				case 1: {
					rd_place->suffix=&sw_info->vlan_egress_suffix;
					break;
				}
				case 2: {
					rd_place->suffix=&sw_info->vlan_forbidden_suffix;
					break;
				}
				case 3: {
					rd_place->suffix=&sw_info->vlan_untagged_suffix;
					break;
				}
				case 4: {
					rd_place->suffix=&sw_info->vlan_name_suffix;
					break;
				}
			}
			if (i!=4) {
				rd_place->next=(row_snmp_data_p)Calloc(1,sizeof(row_snmp_data_s));
				rd_place=rd_place->next;
			}
			else rd_place->next=NULL;
		}
		if (!get_row_data(current->row,sw_info)) {
			err_msg("Unable to retrieve vlan information");
			free(current);
			free(cur);
			free(buf);
			free(vb->oid.buf);
			free(vb);
			free(init);
			return NULL;
		}
		cur->entry=(void*)current;
		cur->type=VLAN_ENTRY;
		cur->flags=0;
		cur->next=(list_entry_p)Malloc(sizeof(list_entry));
		cur=cur->next;
		(*num)++;
		init->len=sizeof(cl_msg_hdr);
		init->vb_n=0;
		vb->oid=newmsg->data->oid;
		copy_place=(char*)(init+1);
		if ((shift=try_fit(&init,vb,copy_place))==0) {
			err_msg("Couldn't construct get-next request because of too little MAX_MSG");
			free(init);
			free(vb->oid.buf);
			free(vb);
			return NULL;
		}
		i=send(sw_info->socket, init,init->len,0);
		buf=(char *)Malloc(MAX_MSG);
		i=recvfrom(sw_info->socket, buf, MAX_MSG,0,NULL,NULL);
		if ((newmsg=proceed_request(buf,i,sw_info->mygvp))==NULL) {
			free(buf);
			free(vb->oid.buf);
			free(vb);
			free(init);
			err_msg("Unable to parse responce message");
			return NULL;
		}
	}
	return cur;	
}


int
get_row_data(row_snmp_p target,sw_inf_p sw_info){
	row_snmp_data_p rd_place;
	vb_data_p vb;
	char * copy_place;
	mesg_data_p newmsg;
	char *buf;
	OBJECT_IDENTIFIER_t tmp;
	int shift;
	int i;
	cl_msg_hdr_p init;

	init=(cl_msg_hdr_p)Calloc(1,sizeof(char)*MAX_MSG);
	init->len=sizeof(cl_msg_hdr);
	bcopy(&(sw_info->ip),&(init->ip),sizeof(struct in_addr));
	init->pdu_pr=PDUs_PR_get_request;
	init->vb_n=0;
	rd_place=target->row_data;
	copy_place=(char*)(init+1);
	vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
        vb->vbv_present=VarBindValue_PR_unSpecified;
	vb->os_present=0;
	vb->st_present=0;
	vb->value.size=0;
	while (rd_place) {
		concatenate_oid(&(tmp),target->row_begin,rd_place->suffix);
		concatenate_oid(&(vb->oid),&tmp,&(target->tag_row_id));
		free(tmp.buf);
		if ((shift=try_fit(&init,vb,copy_place))==0) {
			err_msg("Couldn't construct get-next request because of too little MAX_MSG");
			free(init);
			free(vb->oid.buf);
			free(vb);
			return NULL;
		}
		copy_place+=shift;
		rd_place=rd_place->next;
	}
	i=send(sw_info->socket, init,init->len,0);
    	buf=(char *)Malloc(MAX_MSG);
    	i=recvfrom(sw_info->socket, buf, MAX_MSG,0,NULL,NULL);
    	if ((newmsg=proceed_request(buf,i,sw_info->mygvp))==NULL || newmsg->error_status) {
		free(buf);
		free(vb->oid.buf);
		free(vb);
		free(init);
		err_msg("Unable to parse responce message");
		return 0;
	}
	rd_place=target->row_data;
	vb=newmsg->data;
	while (rd_place) {
		rd_place->data=vb++;
		rd_place=rd_place->next;
	}
	free(newmsg->community.buf);
	free(newmsg);
	free(buf);
	free(init);
	return 1;
}

int
set_row_data(row_snmp_p target,sw_inf_p sw_info){
	row_snmp_data_p rd_place;
	vb_data_p vb;
	char * copy_place;
	mesg_data_p newmsg;
	char *buf;
	OBJECT_IDENTIFIER_t tmp;
	int shift;
	int i,j;
	cl_msg_hdr_p init;
	cl_vb_hdr_p  vb_init;

	init=(cl_msg_hdr_p)Calloc(1,sizeof(char)*MAX_MSG);
	init->len=sizeof(cl_msg_hdr);
	bcopy(&(sw_info->ip),&(init->ip),sizeof(struct in_addr));
	init->pdu_pr=PDUs_PR_set_request;
	init->vb_n=0;
	rd_place=target->row_data;
	copy_place=(char*)(init+1);
	vb=(vb_data_p)Calloc(1,sizeof(vb_data_s));
	while (rd_place) {
		vb->vbv_present=rd_place->data->vbv_present;
		vb->os_present=rd_place->data->os_present;
		vb->st_present=rd_place->data->st_present;
		concatenate_oid(&(tmp),target->row_begin,rd_place->suffix);
		concatenate_oid(&(vb->oid),&tmp,&(target->tag_row_id));
		free(tmp.buf);
		vb->value=rd_place->data->value;
		if ((shift=try_fit(&init,vb,copy_place))==0) {
			err_msg("Couldn't construct get-next request because of too little MAX_MSG");
			free(init);
			free(vb->oid.buf);
			free(vb);
			return 0;
		}
		copy_place+=shift;
		rd_place=rd_place->next;
	}
	if (!sw_info->readonly){
		i=send(sw_info->socket, init,init->len,0);
		sw_info->reconfigure=1;
    		buf=(char *)Malloc(MAX_MSG);
    		i=recvfrom(sw_info->socket, buf, MAX_MSG,0,NULL,NULL);
    		if ((newmsg=proceed_request(buf,i,sw_info->mygvp))==NULL || newmsg->error_status) {
			free(buf);
			free(vb->oid.buf);
			free(vb);
			free(init);
			err_msg("Unable to parse responce message");
			return 0;
		}
		free(newmsg->community.buf);
		free(newmsg);
		free(buf);
	}
	else {
		printf("Will send set request for ");
		copy_place=(char*)(init+1);
		for (i=0;i<init->vb_n;i++){
			vb_init=(cl_vb_hdr_p)copy_place;
			copy_place=(char*)(vb_init+1);
			tmp.size=vb_init->oid_len;
			tmp.buf=(uint8_t *)Malloc(tmp.size);
			bcopy(copy_place,tmp.buf,tmp.size);
			print_arcs(&tmp);
			free(tmp.buf);
			copy_place=(char*)(vb_init+1)+vb_init->oid_len;
			printf(" arg type %i of size %i of contents " ,vb_init->flags,vb_init->len-vb_init->oid_len-sizeof(cl_vb_hdr));
			for (j=0;j<vb_init->len-vb_init->oid_len-sizeof(cl_vb_hdr);j++)
				printf("%u ",copy_place[j]);
			printf("\n");
			copy_place=(char *)(vb_init) + vb_init->len;
		}
	}
	free(init);
	return 1;
}		


int
concatenate_oid(OBJECT_IDENTIFIER_t *oid,OBJECT_IDENTIFIER_t *first,OBJECT_IDENTIFIER_t *second) { 
	oid->size=first->size+second->size;
	oid->buf=(uint8_t *)Malloc(sizeof(uint8_t)*oid->size);
	bcopy(first->buf,oid->buf,first->size);
	bcopy(second->buf,&(oid->buf[first->size]),second->size);
	return 0;
}


l_ref_p
get_current(sw_inf_p sw_info,int entries){
	int i;
	int j=0;
	list_entry_p cur,beg,tmp;
	int number=0;
	int num=0;
	l_ref_p res;
	
	cur=(list_entry_p)Malloc(sizeof(list_entry));
	beg=cur;
	for (i=entries;i;i>>=1) {
		if ((i & 1) && (par_hdlrs[j].c_p_get_f)){
			if ((tmp=(*(par_hdlrs[j]).c_p_get_f)(&cur,sw_info,&num))!=NULL){
				cur=tmp;
				number+=num;
				num=0;
			}	
		}
		j++;
	}
	cur->type=END_OF_LIST;
	cur->next=NULL;
	cur->flags=0;
	cur->entry=NULL;
	number++;
	
	res=(l_ref_p)Malloc(sizeof(l_ref));
	res->base=(list_entry_p_base)Malloc(sizeof(list_entry_p)*number);
	res->number=number;
	res->entries=entries;
	cur=beg;
	for (i=0;i<number;i++) {
		(res->base[i])=cur;
		cur=cur->next;
		if ((cur==NULL) && (i<number-1))
			err_msg("Wrong pointer");
	}
	qsort((void *)res->base, res->number, sizeof(list_entry_p), (int (*)(const void *, const void *))compare_entry);
	for (i=1;i<number; i++) {
		if (compare_entry(&(res->base[i]),&(res->base[i-1]))==0) {
			(res->base[i])->flags |= NOT_PERFORM;
			err_msg("Received duplicate current config. Investigate");	
		}
	}
	return res;
}

void
usage(){
	printf("swconf [-r] ip_switch_to_be_configured\n");
	exit(-1);
}


int
main(int argc, char *argv[]){
	sw_inf_p	sw_info;
	l_ref_p		t_list,c_list;
	int		i,j,k;
	int		res;
	int		type_beg=0;
	int		ch;

	struct sockaddr_un *serv;
	const char *path_to_socket="/tmp/ctrlsnmpd_socket";
	char *comm="private"; /*Should also update commlen*/
	int commlen=7;
	int type;


	sw_info=(sw_inf_p)Malloc(sizeof(sw_inf));
	sw_info->readonly=0;
	/*Parse command line options*/	
	while ((ch = getopt(argc,argv,"r")) != -1) {
		switch (ch) {
			case 'r':
				sw_info->readonly++;
				break;
			case '?':
			default:
				usage();
		}
	}
	
        argc -=optind;
        argv +=optind;
	if (!argc)
		usage();
	sw_info->num_ports=26;
	oid_create(&sw_info->vlan_row_oid,base_of_tree);
	inet_aton(argv[0],&(sw_info->ip));
	sw_info->socket=Socket(AF_LOCAL,SOCK_STREAM,0);
	serv=(struct sockaddr_un *)Calloc(1,sizeof(struct sockaddr_un));
	serv->sun_family=AF_LOCAL;
	strncpy(serv->sun_path,path_to_socket,strlen(path_to_socket));
	Connect(sw_info->socket,(struct sockaddr *)serv,sizeof(struct sockaddr_un));

	sw_info->mygvp=(mgvp)Malloc(sizeof(mgv));
	sw_info->mygvp->version=1;
	sw_info->mygvp->community.buf=(uint8_t *)Malloc(sizeof(char)*commlen);
    	if ((strncpy(sw_info->mygvp->community.buf,comm,commlen))==NULL)
		err_quit("Unable to copy default community");
	sw_info->mygvp->community.size=commlen;

	sw_info->vlan_name_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->vlan_name_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->vlan_name_suffix.buf, &(name_suffix), sizeof(long),0);
	sw_info->vlan_egress_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->vlan_egress_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->vlan_egress_suffix.buf, &(egress_suffix), sizeof(long),0);
	sw_info->vlan_forbidden_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->vlan_forbidden_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->vlan_forbidden_suffix.buf, &(forbidden_suffix), sizeof(long),0);
	sw_info->vlan_untagged_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->vlan_untagged_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->vlan_untagged_suffix.buf, &(untagged_suffix), sizeof(long),0);
	sw_info->vlan_init_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->vlan_init_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->vlan_init_suffix.buf, &(init_suffix), sizeof(long),0);

	sw_info->vlan_untagged_ports.size=sw_info->num_ports/8+1;
	sw_info->vlan_untagged_ports.buf=(uint8_t *)Calloc(1,sw_info->num_ports/8+1);

	oid_create(&sw_info->port_row_oid,base_of_tree_ports);
	sw_info->port_state_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->port_state_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->port_state_suffix.buf, &(port_state_suffix), sizeof(long),0);
	sw_info->port_speed_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->port_speed_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->port_speed_suffix.buf, &(port_speed_suffix), sizeof(long),0);
	sw_info->port_index_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->port_index_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->port_index_suffix.buf, &(port_index_suffix), sizeof(long),0);

	oid_create(&sw_info->fdb_st_row_oid,base_of_tree_fdb_st);
	sw_info->fdb_st_port_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->fdb_st_port_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->fdb_st_port_suffix.buf, &(fdb_st_port_suffix), sizeof(long),0);
	sw_info->fdb_st_state_suffix.buf=(uint8_t *)Malloc((sizeof(long)*8) / 7+1);
	sw_info->fdb_st_state_suffix.size=OBJECT_IDENTIFIER_set_single_arc(sw_info->fdb_st_state_suffix.buf, &(fdb_st_state_suffix), sizeof(long),0);

	sw_info->reconfigure=1;

	if ((t_list=open_task(sw_info))==NULL)
		err_quit("Couldn't get the task");
	while (sw_info->reconfigure) {
		sw_info->reconfigure=0;
		i=j=0;
		if ((c_list=get_current(sw_info,t_list->entries))==NULL)
			err_quit("Couldn't get current configuration");
		if (t_list->base[0]->type > c_list->base[0]->type)
			type=c_list->base[0]->type;
		else
			type=t_list->base[0]->type;
		while (i<t_list->number && j<c_list->number) {
			if (t_list->base[i]->flags & NOT_PERFORM) {
				i++;
				continue;
			}
			if ((t_list->base[i]->type > type) && (c_list->base[j]->type > type)) {
				/*Here we should do post type configuration - such as vlan untagged for not async vlans*/
				if (t_list->base[i]->type > c_list->base[j]->type)
					type=c_list->base[j]->type;
				else
					type=t_list->base[i]->type;
				for (k=type_beg;k<i;k++) {
					if (!entry_post_config(t_list->base[k],sw_info))
						err_quit("Post_function_failed");
				}
				type_beg=i;
			}
			res=compare_entry_no_err(&(t_list->base[i]),&(c_list->base[j]));
			if (t_list->base[i]->type==END_OF_LIST && c_list->base[j]->type==END_OF_LIST){
				i++;j++;
				continue;
			}
			switch (res) {
				case -1:{
					/*need to create parameter*/
					if (!create_entry(t_list->base[i],sw_info))
						err_quit("Create failed");;
					i++;
					break;
				}
				case 0:{
					/*need to check parameter*/
					if (!check_entry(t_list->base[i],c_list->base[j],sw_info))
						err_quit("Check failed");
					
					i++;j++;
					break;
				}
				case 1:{
					/*need to delete parameter*/
					if (!delete_entry(c_list->base[j],sw_info))
						err_quit("Delete vlan failed");
					j++;
					break;
				}
			}
		} 
	}
	return 0;
}


int
create_entry(list_entry_p entry,sw_inf_p sw_info){
	
	if (par_hdlrs[entry->type].c_p_cr_f){
		if (!(*(par_hdlrs[entry->type]).c_p_cr_f)(entry,sw_info)){
			err_msg("Unable perform creation");
			return 0;
		}	
	}
	else {
		err_msg("Function not available");
		return 0;
	}
	return 1;
}

int
delete_entry(list_entry_p entry,sw_inf_p sw_info){

	if (par_hdlrs[entry->type].c_p_del_f){
		if (!(*(par_hdlrs[entry->type]).c_p_del_f)(entry,sw_info)){
			err_msg("Unable perform deletion");
			return 0;
		}	
	}
	else {
		err_msg("Function not available");
		return 0;
	}
	return 1;
}

int
check_entry(list_entry_p t_entry, list_entry_p c_entry,sw_inf_p sw_info){

	if (par_hdlrs[t_entry->type].c_p_chk_f){
		if (!(*(par_hdlrs[t_entry->type]).c_p_chk_f)(t_entry,c_entry,sw_info)){
			err_msg("Unable perform deletion");
			return 0;
		}	
	}
	else {
		err_msg("Function not available");
		return 0;
	}
	return 1;	
} 

int
entry_post_config(list_entry_p entry,sw_inf_p sw_info){
	register const struct post_functions_handler *p;
	int done=0;
	
	for (p=post_funcs;p->p_name;p++){
		if (p->type==entry->type) {
			if (p->flags_to_be_up && entry->flags) {
				done=1;
				if (p->c_p_post_f){
					if (!(*(p->c_p_post_f))(entry,sw_info)){
						err_msg("Post function Failed");
						return 0;
					}
				}
				else {
					err_msg("Function not available");
					return 0;
				} 
			}
		}
	}
	return 1;
}


int
fdb_st_check(list_entry_p t_entry, list_entry_p c_entry, sw_inf_p sw_info){
	fdb_snmp_p target=(fdb_snmp_p)(t_entry->entry);
	fdb_snmp_p current=(fdb_snmp_p)(c_entry->entry);

	row_snmp_p row=NULL;
	row_snmp_data_p c_data,t_data,rd_place,c_runner,t_runner;
	int found;
	OBJECT_IDENTIFIER_t 	*suffix;

	int i,j;

	t_data=target->row->row_data;
	c_data=current->row->row_data;
	found=0;
	suffix=&sw_info->fdb_st_port_suffix;
	for (c_runner=c_data;c_runner;c_runner=c_runner->next) {
		if (c_runner->suffix==suffix) {
			found++;
			break;
		}
	}
	if (!found) {
		err_msg("Couldn't find current value");
		return 0;
	}
	found=0;
	for (t_runner=t_data;t_runner;t_runner=t_runner->next) {
		if (t_runner->suffix==suffix) {
			found++;
			break;
		}
	}
	if (!found) {
		err_msg("Couldn't find task value");
		return 0;
	}
	found=0;
	for (j=0;j<t_runner->data->value.size;j++) {
		if (c_runner->data->value.buf[j] != (c_runner->data->value.buf[j] & t_runner->data->value.buf[j])) {
			found=1;
		}
	}
	if (found) {
		if (!fdb_st_delete(c_entry,sw_info)){
			err_msg("Could not delete fdb in check_fdb");
			return 0;
		}
		if (!fdb_st_create(t_entry,sw_info)){
			err_msg("Couldn't create fdb in check_fdb");
			return 0;
		}
		return 1;
	}
	return 1;
}

int
port_check(list_entry_p t_entry, list_entry_p c_entry, sw_inf_p sw_info){
	port_snmp_p target=(port_snmp_p)(t_entry->entry);
	port_snmp_p current=(port_snmp_p)(c_entry->entry);
	row_snmp_p row=NULL;
	row_snmp_data_p c_data,t_data,rd_place,c_runner,t_runner;
	int i,j;
	int found;
	OBJECT_IDENTIFIER_t 	*suffix;


	t_data=target->row->row_data;
	c_data=current->row->row_data;
	for (i=1;i<3;i++) {
		found=0;
		switch (i) {
			case 1: {
				suffix=&sw_info->port_state_suffix;
				break;
			}
			case 2: {
				suffix=&sw_info->port_speed_suffix;
				break;
			}
		}
		for (c_runner=c_data;c_runner;c_runner=c_runner->next) {
			if (c_runner->suffix==suffix) {
				found++;
				break;
			}
		}
		if (!found) {
			err_msg("Couldn't find current value");
			return 0;
		}
		found=0;
		for (t_runner=t_data;t_runner;t_runner=t_runner->next) {
			if (t_runner->suffix==suffix) {
				found++;
				break;
			}
		}
		if (!found) {
			err_msg("Couldn't find task value");
			return 0;
		}
		if (*((long *)t_runner->data->value.buf)!=*((long *)c_runner->data->value.buf)) {
			if (!row) {
				row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
				row->row_begin=&sw_info->port_row_oid;
				row->tag_row_id.buf=(uint8_t *)Malloc(target->row->tag_row_id.size);
				row->tag_row_id.size=target->row->tag_row_id.size;
				bcopy(target->row->tag_row_id.buf,row->tag_row_id.buf,row->tag_row_id.size);
				row->row_data=rd_place=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
			}
			else {
				rd_place->next=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
				rd_place=rd_place->next;
			}
			rd_place->suffix=suffix;
			rd_place->data=(vb_data_p)Calloc(1,sizeof(vb_data_s));
			rd_place->data->vbv_present=VarBindValue_PR_value;
			rd_place->data->os_present=ObjectSyntax_PR_simple;
			rd_place->data->st_present=SimpleSyntax_PR_integer_value;
			rd_place->next=NULL;
			os_copy(&(rd_place->data->value),&(t_runner->data->value));
		}
	}
	if (row) {
		if (!set_row_data(row,sw_info)) {
			err_msg("Unable to set port information");
			return 0;
		}
	}
	return 1;
}

int
vlan_check(list_entry_p t_entry, list_entry_p c_entry, sw_inf_p sw_info){
	vl_snmp_p target=(vl_snmp_p)(t_entry->entry);
	vl_snmp_p current=(vl_snmp_p)(c_entry->entry);
	row_snmp_p row;
	row_snmp_data_p c_data,t_data,rd_place,c_runner,t_runner;
	int i,j;
	int found,set;
	OCTET_STRING_t	common;
	int number=0;

	row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
	row->row_begin=&sw_info->vlan_row_oid;
	row->tag_row_id.buf=(uint8_t *)Malloc(target->row->tag_row_id.size);
	row->tag_row_id.size=target->row->tag_row_id.size;
	bcopy(target->row->tag_row_id.buf,row->tag_row_id.buf,row->tag_row_id.size);
	row->row_data=rd_place=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
	t_data=target->row->row_data;
	c_data=current->row->row_data;
	common.buf=(uint8_t *)Malloc(c_data->data->value.size);
	common.size=c_data->data->value.size;
	for (i=1;i<4;i++) {
		found=0;set=0;
		switch (i) {
			case 1: {
				/*Delete untagged that do not match*/
				rd_place->suffix=&sw_info->vlan_untagged_suffix;
				break;
			}
			case 2: {
				/*Delete some forbidden ports*/
				rd_place->suffix=&sw_info->vlan_forbidden_suffix;
				break;
			}
			case 3: {
				/*Delete some egress ports*/
				rd_place->suffix=&sw_info->vlan_egress_suffix;
				break;
			}	
		}
		for (c_runner=c_data;c_runner;c_runner=c_runner->next) {
			if (c_runner->suffix==rd_place->suffix) {
				found++;
				break;
			}
		}
		if (!found) {
			err_msg("Couldn't find current untagged string");
			return 0;
		}
		found=0;
		for (t_runner=t_data;t_runner;t_runner=t_runner->next) {
			if (t_runner->suffix==rd_place->suffix) {
				found++;
				break;
			}
		}
		if (!found) {
			err_msg("Couldn't find task untagged string");
			return 0;
		}
		for (j=0;j<t_runner->data->value.size;j++) {
			common.buf[j]=c_runner->data->value.buf[j] & t_runner->data->value.buf[j];
			if (c_runner->data->value.buf[j] != (c_runner->data->value.buf[j] & t_runner->data->value.buf[j])) {
				set=1;
			}
		}
		if (set){
			err_msg("Should delete some %i ports",i);
			rd_place->data=(vb_data_p)Calloc(1,sizeof(vb_data_s));
			rd_place->data->vbv_present=VarBindValue_PR_value;
			rd_place->data->os_present=ObjectSyntax_PR_simple;
			rd_place->data->st_present=SimpleSyntax_PR_string_value;
			os_copy(&(rd_place->data->value),&common);
			rd_place->next=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
			rd_place=rd_place->next;
			number++;
		}
	}
	for (i=1;i<3;i++) {
		found=0;
		switch (i) {
			case 1: {
				/*Delete untagged that do not match*/
				rd_place->suffix=&sw_info->vlan_egress_suffix;
				break;
			}
			case 2: {
				/*Delete some forbidden ports*/
				rd_place->suffix=&sw_info->vlan_forbidden_suffix;
				break;
			}
		}
		for (c_runner=c_data;c_runner;c_runner=c_runner->next) {
			if (c_runner->suffix==rd_place->suffix) {
				found++;
				break;
			}
		}
		if (!found) {
			err_msg("Couldn't find current untagged string");
			return 0;
		}
		found=0;
		for (t_runner=t_data;t_runner;t_runner=t_runner->next) {
			if (t_runner->suffix==rd_place->suffix) {
				found++;
				break;
			}
		}
		if (!found) {
			err_msg("Couldn't find task untagged string");
			return 0;
		}
		set=0;
		for (j=0;j<t_runner->data->value.size;j++) {
			common.buf[j]=t_runner->data->value.buf[j];
			if (t_runner->data->value.buf[j] != (c_runner->data->value.buf[j] & t_runner->data->value.buf[j])) {
				set=1;
			}
		}
		if (set) {
			err_msg("Should add some %i ports",i);
			rd_place->data=(vb_data_p)Calloc(1,sizeof(vb_data_s));
			rd_place->data->vbv_present=VarBindValue_PR_value;
			rd_place->data->os_present=ObjectSyntax_PR_simple;
			rd_place->data->st_present=SimpleSyntax_PR_string_value;
			os_copy(&(rd_place->data->value),&common);
			rd_place->next=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
			rd_place=rd_place->next;
			number++;
		}
	}
	rd_place=row->row_data;
	for (i=0;i<number-1;i++)
		rd_place=rd_place->next;
	free(rd_place->next);
	rd_place->next=NULL;
	if (number) {
		if (!set_row_data(row,sw_info)) {
			err_msg("Unable to set vlan information");
			return 0;
		}
	}
	for (c_runner=c_data;c_runner;c_runner=c_runner->next) {
		if (c_runner->suffix==&sw_info->vlan_untagged_suffix) {
			found++;
			break;
		}
	}
	if (!found) {
		err_msg("Couldn't find current untagged string");
		return 0;
	}
	found=0;
	for (t_runner=t_data;t_runner;t_runner=t_runner->next) {
		if (t_runner->suffix==&sw_info->vlan_untagged_suffix) {
			found++;
			break;
		}
	}
	if (!found) {
		err_msg("Couldn't find task untagged string");
		return 0;
	}
	for (j=0;j<t_runner->data->value.size;j++) {
		if (t_runner->data->value.buf[j] != (c_runner->data->value.buf[j] & t_runner->data->value.buf[j])) {
			err_msg("Should do untag\n");
			t_entry->flags |= DO_UNTAG;
		}
	}
	return 1;
}


int
port_delete(list_entry_p entry,sw_inf_p sw_info){
	err_msg("No port can be deleted");
	return 1;
}

int
port_create(list_entry_p entry,sw_inf_p sw_info){
	err_msg("No port can be created");
	return 1;
}


int
fdb_st_delete(list_entry_p entry,sw_inf_p sw_info){
	fdb_snmp_p target=(fdb_snmp_p)(entry->entry);
	OBJECT_IDENTIFIER_t 	*suffix;
	row_snmp_p row;
	row_snmp_data_p data,rd_place;
	row_snmp_data_p runner;
	int destroy=2;
	int found=0;

	suffix=&sw_info->fdb_st_state_suffix;
	for (runner=target->row->row_data;runner;runner=runner->next) {
		if (runner->suffix==suffix) {
			found++;
			break;
		}
	}
	if (!found) {
			err_msg("Couldn't find index value");
			return 0;
	}
	row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
	row->row_begin=&sw_info->fdb_st_row_oid;
	row->tag_row_id.buf=(uint8_t *)Malloc(target->row->tag_row_id.size);
	row->tag_row_id.size=target->row->tag_row_id.size;
	bcopy(target->row->tag_row_id.buf,row->tag_row_id.buf,row->tag_row_id.size);
	row->row_data=rd_place=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
	rd_place->data=(vb_data_p)Calloc(1,sizeof(vb_data_s));
	rd_place->suffix=&sw_info->fdb_st_state_suffix;
	rd_place->data->vbv_present=VarBindValue_PR_value;
	rd_place->data->os_present=ObjectSyntax_PR_simple;
	rd_place->data->st_present=SimpleSyntax_PR_integer_value;
	rd_place->data->value.size=sizeof(long);
	rd_place->data->value.buf=(uint8_t *)Malloc(rd_place->data->value.size);
	bcopy(&destroy,rd_place->data->value.buf,sizeof(long));
	rd_place->next=NULL;
	if (!set_row_data(row,sw_info)) {
		err_msg("Unable to delete fdb information");
		return 0;
	}
	return 1;
}

int
vlan_delete(list_entry_p entry,sw_inf_p sw_info){
	vl_snmp_p target=(vl_snmp_p)(entry->entry);
	row_snmp_p row;
	row_snmp_data_p data,rd_place;
	int delete_value=6;
	int i;


	if (target->vid==1) {
		err_msg("Couldn't delete vlan default. Just marking all ports as forbidden");
		return 1;
	}
	row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
	row->row_begin=&sw_info->vlan_row_oid;
	row->tag_row_id.buf=(uint8_t *)Malloc(target->row->tag_row_id.size);
	row->tag_row_id.size=target->row->tag_row_id.size;
	bcopy(target->row->tag_row_id.buf,row->tag_row_id.buf,row->tag_row_id.size);
	row->row_data=rd_place=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
	data=target->row->row_data;
	rd_place->data=(vb_data_p)Calloc(1,sizeof(vb_data_s));
	rd_place->suffix=&sw_info->vlan_init_suffix;
	rd_place->data->vbv_present=VarBindValue_PR_value;
	rd_place->data->os_present=ObjectSyntax_PR_simple;
	rd_place->data->st_present=SimpleSyntax_PR_integer_value;
	rd_place->data->value.size=sizeof(long);
	rd_place->data->value.buf=(uint8_t *)Malloc(rd_place->data->value.size);
	bcopy(&delete_value,rd_place->data->value.buf,sizeof(long));
	rd_place->next=NULL;
	if (!set_row_data(row,sw_info)) {
		err_msg("Unable to set vlan information");
		return 0;
	}
	return 1;
}	


int
fdb_st_create(list_entry_p entry,sw_inf_p sw_info){
	fdb_snmp_p target=(fdb_snmp_p)(entry->entry);
	
	if (!set_row_data(target->row,sw_info)) {
		err_msg("Unable to create fdb information");
		return 0;
	}
	return 1;
}

int
vlan_create(list_entry_p entry,sw_inf_p sw_info){
	vl_snmp_p target=(vl_snmp_p)(entry->entry);
	row_snmp_p row;
	row_snmp_data_p data,rd_place,runner;
	int create_value=4;
	int i,found;

	row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
	row->row_begin=&sw_info->vlan_row_oid;
	row->tag_row_id.buf=(uint8_t *)Malloc(target->row->tag_row_id.size);
	row->tag_row_id.size=target->row->tag_row_id.size;
	bcopy(target->row->tag_row_id.buf,row->tag_row_id.buf,row->tag_row_id.size);
	row->row_data=rd_place=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
	data=target->row->row_data;
	for (i=1;i<5;i++) {
		rd_place->data=(vb_data_p)Calloc(1,sizeof(vb_data_s));
		switch (i) {
			case 1: {
				rd_place->suffix=&sw_info->vlan_egress_suffix;
				break;
			}
			case 2: {
				rd_place->suffix=&sw_info->vlan_forbidden_suffix;
				break;
			}
			case 3: {
				rd_place->suffix=&sw_info->vlan_name_suffix;
				break;
			}
			case 4: {
				rd_place->suffix=&sw_info->vlan_init_suffix;
				rd_place->data->vbv_present=VarBindValue_PR_value;
				rd_place->data->os_present=ObjectSyntax_PR_simple;
				rd_place->data->st_present=SimpleSyntax_PR_integer_value;
				rd_place->data->value.size=sizeof(long);
				rd_place->data->value.buf=(uint8_t *)Malloc(rd_place->data->value.size);
				bcopy(&create_value,rd_place->data->value.buf,sizeof(long));
				break;
			}
		}
		if (i!=4) {
			found=0;
			for (runner=data;runner;runner=runner->next) {
					if (runner->suffix==rd_place->suffix) {
						found=1;
						break;
					}
			}
			if (found) {
				rd_place->data->vbv_present=VarBindValue_PR_value;
				rd_place->data->os_present=ObjectSyntax_PR_simple;
				rd_place->data->st_present=SimpleSyntax_PR_string_value;
				rd_place->data->value.size=runner->data->value.size;
				rd_place->data->value.buf=(uint8_t *)Malloc(rd_place->data->value.size);
				bcopy(runner->data->value.buf,rd_place->data->value.buf,rd_place->data->value.size);
			}
			else {
				err_msg("Couldn't find current string");
				return 0;
			}
			rd_place->next=(row_snmp_data_p)Calloc(1,sizeof(row_snmp_data_s));
			rd_place=rd_place->next;
		}
		else rd_place->next=NULL;
	}
	found=0;
	for (runner=data;runner;runner=runner->next) {
		if (runner->suffix==&sw_info->vlan_untagged_suffix) {
			found=1;
			break;
		}
	}
	if (found) {
		for (i=0;i<runner->data->value.size;i++){
			if (runner->data->value.buf[i] !=0) {
				entry->flags |= DO_UNTAG;
				err_msg("Should do untag after creation");
				break;
			}
		}
	}
	else {
		err_msg("Couldn't find current string");
		return 0;
	}
	if (!set_row_data(row,sw_info)) {
		err_msg("Unable to set vlan information");
		return 0;
	}
	return 1;
}

int
vlan_untag(list_entry_p entry,sw_inf_p sw_info){
	vl_snmp_p target=(vl_snmp_p)(entry->entry);
	row_snmp_p row;
	row_snmp_data_p data,rd_place,runner;
	int found=0;
	
	row=(row_snmp_p)Malloc(sizeof(row_snmp_s));
	row->row_begin=&sw_info->vlan_row_oid;
	row->tag_row_id.buf=(uint8_t *)Malloc(target->row->tag_row_id.size);
	row->tag_row_id.size=target->row->tag_row_id.size;
	bcopy(target->row->tag_row_id.buf,row->tag_row_id.buf,row->tag_row_id.size);
	row->row_data=rd_place=(row_snmp_data_p)Malloc(sizeof(row_snmp_data_s));
	data=target->row->row_data;
	rd_place->data=(vb_data_p)Calloc(1,sizeof(vb_data_s));
	rd_place->suffix=&sw_info->vlan_untagged_suffix;
	for (runner=data;runner;runner=runner->next) {
		if (runner->suffix==&sw_info->vlan_untagged_suffix) {
			found=1;
			break;
		}
	}
	if (found) {
		rd_place->data->vbv_present=VarBindValue_PR_value;
		rd_place->data->os_present=ObjectSyntax_PR_simple;
		rd_place->data->st_present=SimpleSyntax_PR_string_value;
		rd_place->data->value.size=runner->data->value.size;
		rd_place->data->value.buf=(uint8_t *)Malloc(rd_place->data->value.size);
		bcopy(runner->data->value.buf,rd_place->data->value.buf,rd_place->data->value.size);
		rd_place->next=NULL;
	}
	else {
		err_msg("Couldn't find current string");
		return 0;
	}
	if (!set_row_data(row,sw_info)) {
		err_msg("Unable to set vlan information");
		return 0;
	}
	return 1;
}	

int
print_arcs(OBJECT_IDENTIFIER_t *oid) {
		unsigned long fixed_arcs[10];	
		unsigned long *arcs = fixed_arcs;
		int arc_type_size = sizeof(fixed_arcs[0]);	
		int arc_slots = sizeof(fixed_arcs)/sizeof(fixed_arcs[0]);
		int count;	
		int i;

		count = OBJECT_IDENTIFIER_get_arcs(oid, arcs,
			arc_type_size, arc_slots);
		if(count > arc_slots) {
			arc_slots = count;
			arcs = malloc(arc_type_size * arc_slots);
			if  (!arcs) 
			    return(-1);
			count = OBJECT_IDENTIFIER_get_arcs(oid, arcs,
				arc_type_size, arc_slots);
			if  (count != arc_slots)
			    return(-1);
		}
 		for(i = 0; i < count; i++) {
			printf(".%lu", arcs[i]);
		}
		if(arcs != fixed_arcs) free(arcs);
		return 0;
}
